{-# LANGUAGE DeriveDataTypeable, RecordWildCards #-}
{-# OPTIONS -Wall #-}

-- | Main entry point.

module Main where

import           Control.Concurrent
import           Control.Monad
import           Data.Data               (Data,Typeable)
import           Network                 (PortID(PortNumber),listenOn)
import           Network.Socket hiding (listen,recv,send)
import           Network.Socket.ByteString (recv,sendAll)
import           System.Console.CmdArgs  (cmdArgs,(&=),help,summary,def,opt)
import           System.Posix            (Handler(Ignore),installHandler,sigPIPE)

data Throttle = Throttle
  { listen  :: Int
  , host    :: String
  , port    :: Int
  , speed   :: Float
  , logging :: Bool
  } deriving (Show,Data,Typeable)

options :: Throttle
options = Throttle
  { speed   = def &= opt (1.6::Float) &= help "Speed in KB/s, e.g. 1.6 (0 for no limit)"
  , host    = "127.0.0.1"
  , port    = 80
  , listen  = 8000
  , logging = def &= help "Log about events on the console."
  }
  &= summary "Throttle v1.0, (C) Chris Done 2010"
  &= help "Listens on port <listen> and proxies a throttled \
          \connection to <host> on <port> at speed <speed>KB/s."

main :: IO ()
main = do
  void $ installHandler sigPIPE Ignore Nothing
  cmdArgs options >>= start

start :: Throttle -> IO ()
start Throttle{..} = withSocketsDo $ do
  c <- newTeller logging
  listener <- listenOn (PortNumber . fromIntegral $ listen)
  forever $ do
    (client,_) <- accept listener
    tell c $ [show client,": New connection on port ",show port]
    void $ forkIO $ do
      server <- connectToServer
      tell c $ [show client,": ",show server,": Connected to server at "
               ,host,":",show port]
      let proxyTo = proxyToWithChan c
      client `proxyTo` server
      server `proxyTo` client
    return ()
  where connectToServer = do
          addrinfos <- getAddrInfo Nothing (Just host) (Just $ show port)
          let serveraddr = head addrinfos
          server <- socket (addrFamily serveraddr) Stream defaultProtocol
          connect server (addrAddress serveraddr)
          return server
        proxyToWithChan c from to = do
          void $ forkIO $ flip catch (close c from to) $ forever $ do
            payload <- recv from 4096
            sendAll to payload
            when (speed > 0) $ threadDelay delay
          return ()
        close c a b _ = do
          tell c $ [show a,":",show b,": Closing connections."]
          sClose a
          sClose b
        delay = round $ (1000 * 1000) / (speed/4)

-- | Create a new console logger.
newTeller :: Bool -> IO (Maybe (Chan String))
newTeller False = return Nothing
newTeller True = do
  c <- newChan
  void $ forkIO $ do
    getChanContents c >>= mapM_ putStrLn
  return $ Just c

-- | Tell the user something on the console.
tell :: Maybe (Chan String) -> [String] -> IO ()
tell (Just c) = writeChan c . concat
tell Nothing  = const $ return ()
